# -*- coding: utf-8 -*-

"""CLI utils."""

import json
import os
from collections import OrderedDict
import pandas as pd
import click
from pykeen.constants import KG_EMBEDDING_MODEL_NAME


def get_config_dict(model_name):
    """Get configuration dictionary.

    :param str model_name:
    :rtype: dict
    """
    config = OrderedDict()
    config[KG_EMBEDDING_MODEL_NAME] = model_name
    return config


def summarize_results(directory: str, output):
    """Summarize contents of training and evaluation"""
    r = []
    for subdirectory_name in os.listdir(directory):
        subdirectory = os.path.join(directory, subdirectory_name)
        if not os.path.isdir(subdirectory):
            continue
        configuration_path = os.path.join(subdirectory, 'configuration.json')
        if not os.path.exists(configuration_path):
            click.echo("missing configuration")
            continue
        with open(configuration_path) as file:
            configuration = json.load(file)
        evaluation_path = os.path.join(subdirectory, 'evaluation_summary.json')
        if not os.path.exists(evaluation_path):
            click.echo("missing evaluation summary")
            continue
        with open(evaluation_path) as file:
            evaluation = json.load(file)
        r.append(dict(**configuration, **evaluation))
    df = pd.DataFrame(r)
    df.to_csv(output, sep='\t')



# -*- coding: utf-8 -*-

"""PyKEEN is a package for training and evaluating knowledge graph embedding.

Installation
------------
Get the Latest
~~~~~~~~~~~~~~~
Download the most recent code from ... with:

For Developers
~~~~~~~~~~~~~~
Clone the repository from `GitHub <https://github.com/SmartDataAnalytics/pykeen.git>'
"""

from .run import run

__version__ = '0.0.11'

__title__ = 'pykeen'
__description__ = "A package for training and evaluating knowledge graph embeddings"
__url__ = 'https://github.com/SmartDataAnalytics/PyKEEN.git'
__author__ = 'Mehdi Ali'
__email__ = 'mehdi.ali@cs.uni-bonn.de'
__license__ = 'MIT License'
__copyright__ = 'Copyright (c) 2018 Mehdi Ali'



# -*- coding: utf-8 -*-

"""Implementation of the basic pipeline."""

import logging
from typing import Optional

import numpy as np
import rdflib
import torch
from sklearn.model_selection import train_test_split

from pykeen.constants import *
from pykeen.hyper_parameter_optimizer.random_search_optimizer import RandomSearchHPO
from pykeen.utilities.evaluation_utils.metrics_computations import compute_metrics
from pykeen.utilities.initialization_utils.module_initialization_utils import get_kg_embedding_model
from pykeen.utilities.train_utils import train_model
from pykeen.utilities.triples_creation_utils.instance_creation_utils import create_mapped_triples, create_mappings

__all__ = ['Pipeline']

log = logging.getLogger(__name__)


class Pipeline(object):
    def __init__(self, config):
        self.config = config
        self.seed = config[SEED]
        self.entity_to_id = None
        self.rel_to_id = None
        self.device_name = (
            'cuda:0'
            if torch.cuda.is_available() and self.config[PREFERRED_DEVICE] == GPU else
            CPU
        )
        self.device = torch.device(self.device_name)

    def start(self, path_to_train_data: Optional[str] = None):
        is_hpo_mode = self.config[EXECUTION_MODE] == HPO_MODE
        return self._start_pipeline(is_hpo_mode=is_hpo_mode, path_to_train_data=path_to_train_data)

    @property
    def is_evaluation_required(self) -> bool:
        return TEST_SET_PATH in self.config or TEST_SET_RATIO in self.config

    # TODO: Remove path_to_train_data
    def _start_pipeline(self, is_hpo_mode: bool, path_to_train_data: Optional[str] = None):
        pipeline_outcome = OrderedDict()

        if is_hpo_mode:
            mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()

            (trained_model,
             loss_per_epoch,
             entity_to_embedding,
             relation_to_embedding,
             eval_summary,
             params) = RandomSearchHPO.run(
                mapped_train_tripels=mapped_pos_train_triples,
                mapped_test_tripels=mapped_pos_test_triples,
                entity_to_id=self.entity_to_id,
                rel_to_id=self.rel_to_id,
                config=self.config,
                device=self.device,
                seed=self.seed,
            )

        else:

            # Training Mode
            if self.is_evaluation_required:
                mapped_pos_train_triples, mapped_pos_test_triples = self._get_train_and_test_triples()
            else:
                mapped_pos_train_triples = self._get_train_triples()

            all_entities = np.array(list(self.entity_to_id.values()))

            # Initialize KG embedding model
            self.config[NUM_ENTITIES] = len(self.entity_to_id)
            self.config[NUM_RELATIONS] = len(self.rel_to_id)
            self.config[PREFERRED_DEVICE] = CPU if self.device_name == CPU else GPU
            kg_embedding_model = get_kg_embedding_model(config=self.config)

            batch_size = self.config[BATCH_SIZE]
            num_epochs = self.config[NUM_EPOCHS]
            learning_rate = self.config[LEARNING_RATE]
            params = self.config

            log.info("-------------Train KG Embeddings-------------")
            trained_model, loss_per_epoch = train_model(kg_embedding_model=kg_embedding_model,
                                                        all_entities=all_entities,
                                                        learning_rate=learning_rate,
                                                        num_epochs=num_epochs,
                                                        batch_size=batch_size,
                                                        pos_triples=mapped_pos_train_triples,
                                                        device=self.device,
                                                        seed=self.seed)

            eval_summary = None

            if self.is_evaluation_required:
                log.info("-------------Start Evaluation-------------")

                eval_summary = OrderedDict()
                mean_rank, hits_at_k = compute_metrics(all_entities=all_entities,
                                                       kg_embedding_model=kg_embedding_model,
                                                       mapped_train_triples=mapped_pos_train_triples,
                                                       mapped_test_triples=mapped_pos_test_triples,
                                                       device=self.device,
                                                       filter_neg_triples=self.config[FILTER_NEG_TRIPLES])

                eval_summary[MEAN_RANK] = mean_rank
                eval_summary[HITS_AT_K] = hits_at_k

        # Prepare Output
        id_to_entity = {value: key for key, value in self.entity_to_id.items()}
        id_to_rel = {value: key for key, value in self.rel_to_id.items()}
        entity_to_embedding = {
            id_to_entity[id]: embedding.detach().cpu().numpy()
            for id, embedding in enumerate(trained_model.entity_embeddings.weight)
        }

        if self.config[KG_EMBEDDING_MODEL_NAME] in [SE_NAME, UM_NAME]:
            relation_to_embedding = None
        else:
            relation_to_embedding = {
                id_to_rel[id]: embedding.detach().cpu().numpy()
                for id, embedding in enumerate(trained_model.relation_embeddings.weight)
            }

        pipeline_outcome[TRAINED_MODEL] = trained_model
        pipeline_outcome[LOSSES] = loss_per_epoch
        pipeline_outcome[ENTITY_TO_EMBEDDING] = entity_to_embedding
        pipeline_outcome[RELATION_TO_EMBEDDING] = relation_to_embedding
        pipeline_outcome[EVAL_SUMMARY] = eval_summary
        pipeline_outcome[ENTITY_TO_ID] = self.entity_to_id
        pipeline_outcome[RELATION_TO_ID] = self.rel_to_id

        return pipeline_outcome, params

    def _get_train_and_test_triples(self):

        pos_triples = _load_data(self.config[TRAINING_SET_PATH])

        if TEST_SET_PATH in self.config:
            train_pos = pos_triples
            test_pos = _load_data(path_to_data=self.config[TEST_SET_PATH])

        else:
            train_pos, test_pos = train_test_split(
                pos_triples,
                test_size=self.config[TEST_SET_RATIO],
                random_state=self.seed,
            )

        all_triples = np.concatenate([train_pos, test_pos], axis=0)
        self.entity_to_id, self.rel_to_id = create_mappings(triples=all_triples)
        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        mapped_pos_test_triples, _, _ = create_mapped_triples(
            triples=test_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        return mapped_pos_train_triples, mapped_pos_test_triples

    def _get_train_triples(self):
        train_pos = _load_data(self.config[TRAINING_SET_PATH])

        self.entity_to_id, self.rel_to_id = create_mappings(triples=train_pos)

        mapped_pos_train_triples, _, _ = create_mapped_triples(
            triples=train_pos,
            entity_to_id=self.entity_to_id,
            rel_to_id=self.rel_to_id,
        )

        return mapped_pos_train_triples


def _extract_tripels_from_graph(graph):
    """

    :param graph:
    :return:
    """
    triples = []

    for s, p, o in graph:
        triples.append([str(s), str(p), str(o)])

    triples = np.array(triples, dtype=np.str)

    return triples


def _load_data(path_to_data: str):
    try:
        g = rdflib.Graph()
        g.parse(path_to_data, format="nt")
        triples = _extract_tripels_from_graph(g)

    except:
        try:
            triples = np.loadtxt(
                fname=path_to_data,
                dtype=str,
                comments='@Comment@ Subject Predicate Object',
                delimiter='\t',
            )
        except:
            raise ValueError('Dataset must be either a .tsv file of .nt (serilaized rdf) file')

    return triples




import numpy as np

def my_func(r):
    return  hash(tuple(r))
if __name__ == '__main__':
    training_triples = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    test_triples = np.array([[3,3,3], [5, 5, 5], [5, 5, 4]])


    t = np.isin(element=test_triples,test_elements=training_triples) * 1.
    t = np.sum(t,axis=1)/3
    r = np.where(t==1.)
    print(test_triples[r])
    print(t)
    print(r)
    exit(0)
    all_triples = np.concatenate([training_triples,test_triples],axis=0)
    all_triples_hashed = np.apply_along_axis(my_func, 1, all_triples)
    corrupted = np.array([[1,1,1], [1,1,1]])
    corrupted_hashed = np.apply_along_axis(my_func,1,corrupted)

    mask = np.in1d(corrupted_hashed, all_triples_hashed, invert=True)
    mask2 = np.where(mask)[0]

    print(mask2.size)

    if not mask2:
        print("no")

    print(corrupted_hashed[mask])
    print(corrupted_hashed)
    print(corrupted[mask2])




import torch
import torch.autograd
import torch.nn as nn

if __name__ == '__main__':
    pos_scores = torch.tensor([-7.9980, -7.0185, -8.3990], dtype=torch.float)
    neg_scores = torch.tensor([-7.9980, -7.9268, -6.7853], dtype=torch.float)


    criterion = nn.MarginRankingLoss(margin=1, size_average=True)
    y = torch.tensor([1,1,1], dtype=torch.float)
    loss = criterion(pos_scores, neg_scores, y)

    print(loss)



import numpy as np

if __name__ == '__main__':
    a = np.array([5,2,1], dtype=np.float)

    appended = np.append(arr=a, values=7)

    indice_of_pos = appended.size -1

    sorted = np.argsort(appended)

    rank = np.where(sorted == indice_of_pos)[0][0]

    print(a)
    print(appended)
    print(sorted)
    print(indice_of_pos)
    print(rank)


# -*- coding: utf-8 -*-

'''Script for predicting links based on a trained model.'''
import json
import os

import numpy as np
import torch
from pykeen.constants import PREFERRED_DEVICE, GPU, CPU
from pykeen.utilities.initialization_utils.module_initialization_utils import get_kg_embedding_model
from pykeen.utilities.prediction_utils import make_predictions


def start_predictions_piepline(model_direc, data_direc):
    # Load configuration file
    in_path = os.path.join(model_direc, 'configuration.json')
    with open(in_path) as f:
        config = json.load(f)

    # Load entity to id mapping
    in_path = os.path.join(model_direc, 'entity_to_id.json')
    with open(in_path) as f:
        entity_to_id = json.load(f)

    # Load relation to id mapping
    in_path = os.path.join(model_direc, 'relation_to_id.json')
    with open(in_path) as f:
        relation_to_id = json.load(f)

    trained_model = get_kg_embedding_model(config=config)
    path_to_model = os.path.join(model_direc, 'trained_model.pkl')
    trained_model.load_state_dict(torch.load(path_to_model))

    in_path = os.path.join(data_direc, 'entities.tsv')
    entities = np.loadtxt(fname=in_path, dtype=str)

    in_path = os.path.join(data_direc, 'relations.tsv')
    relations = np.loadtxt(fname=in_path, dtype=str)

    device_name = 'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU

    device = torch.device(device_name)

    ranked_triples = make_predictions(kg_model=trained_model,
                                      entities=entities,
                                      relations=relations,
                                      entity_to_id=entity_to_id,
                                      rel_to_id=relation_to_id,
                                      device=device)

    out_path = os.path.join(data_direc, 'predictions.tsv')
    np.savetxt(out_path, ranked_triples, fmt='%s')




import torch
import torch.nn as nn

if __name__ == '__main__':
    emb = torch.nn.Embedding(4, 2)
    print(emb.weight.data)
    norms = torch.norm(emb.weight, p=2, dim=1).data
    emb.weight.data = emb.weight.data.div(norms.view(4, 1).expand_as(emb.weight))

    print()
    print(emb.weight.data)

    print()

    nn.init.uniform_(emb.weight.data, a=-5, b=-3)

    print(emb.weight.data)


import torch
import numpy as np
import torch.nn as nn

if __name__ == '__main__':
    pos_heads = torch.tensor([0, 1])
    pos_rels = torch.tensor([0, 0])
    pos_tails = torch.tensor([0, 0])

    num_entities = 2
    num_relations = 2
    relation_embedding_dim = 3
    entity_embedding_dim = 2

    entity_embeddings = nn.Embedding(num_entities, entity_embedding_dim)
    relation_embeddings = nn.Embedding(num_relations, relation_embedding_dim)

    heads = np.array([[1.,1.,1.],[2.,2.,2.]])
    relations = np.array([[2., 2., 2.], [2., 2., 2.]])
    tails = np.array([[3., 3., 3.], [2., 2., 2.]])

    heads = torch.tensor(heads,dtype=torch.float)
    relations = torch.tensor(relations,dtype=torch.float)
    tails = torch.tensor(tails,dtype=torch.float)

    scores = torch.mul(heads,relations)

    s = torch.sum(heads * relations * tails,dim=1)

    intermediates = torch.mul(heads, relations)
    s2 = torch.einsum('nd,nd->n', [intermediates, tails])

    print(scores)
    print()
    print(s)
    print()
    print(s2)


# -*- coding: utf-8 -*-

"""Implementations of various knowledge graph embedding models."""

from pykeen.kg_embeddings_model.conv_e import ConvE
from pykeen.kg_embeddings_model.distmult import DistMult
from pykeen.kg_embeddings_model.ermlp import ERMLP
from pykeen.kg_embeddings_model.rescal import RESCAL
from pykeen.kg_embeddings_model.structured_embedding import StructuredEmbedding
from pykeen.kg_embeddings_model.trans_d import TransD
from pykeen.kg_embeddings_model.trans_e import TransE
from pykeen.kg_embeddings_model.trans_h import TransH
from pykeen.kg_embeddings_model.trans_r import TransR
from pykeen.kg_embeddings_model.unstructured_model import UnstructuredModel




fname = '/Users/mehdi/PycharmProjects/PyKEEN/data/train.txt'
fname2 = '/Users/mehdi/PycharmProjects/PyKEEN/data/test.txt'

with open(fname) as f:
    content = f.readlines()

with open(fname2) as f:
    content2 = f.readlines()

l = [item for item in content2 if item in content]

print(len(l))



import torch
import numpy as np
import torch.nn as nn

if __name__ == '__main__':
    # a = torch.randn(2, 3)
    # b = torch.randn(2, 3, 7)
    # c = torch.randn(2, 7)
    # torch.einsum('ik,ikl->il', [a, b])
    # exit(0)

    pos_heads = torch.tensor([0,1])
    pos_rels = torch.tensor([0, 0])
    pos_tails = torch.tensor([0, 0])

    num_entities = 2
    num_relations = 2
    relation_embedding_dim = 3
    entity_embedding_dim = 2

    entity_embeddings = nn.Embedding(num_entities, entity_embedding_dim)
    relation_embeddings = nn.Embedding(num_relations, relation_embedding_dim)
    projection_matrix_embs = nn.Embedding(num_relations, entity_embedding_dim * relation_embedding_dim)

    print("projection_matrix_embs shape: ", projection_matrix_embs.weight.shape)
    print("projection_matrix_embs: ", projection_matrix_embs.weight.data)
    print()

    proj_matrix_embs = projection_matrix_embs(pos_rels).view(-1,  entity_embedding_dim, relation_embedding_dim)
    print("proj_matrix_embs shape: ", proj_matrix_embs.shape)
    print(proj_matrix_embs)
    print()

    pos_h_embs = entity_embeddings(pos_heads)
    pos_r_embs = relation_embeddings(pos_rels)
    pos_t_embs = entity_embeddings(pos_tails)

    print("pos_h_embs shape: ", pos_h_embs.shape)
    print("pos_h_embs: ", pos_h_embs)
    print()

    # proj_pos_heads_embs = torch.mm(pos_h_embs, projection_matrix_embs.weight)
    proj_pos_heads_embs = torch.einsum('ik,ikl->il',[pos_h_embs,proj_matrix_embs])
    print("proj_pos_heads_embs shape: ", proj_pos_heads_embs.shape)
    print("proj_pos_heads_embs: ", proj_pos_heads_embs)
    print()






import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim

if __name__ == '__main__':
    entities_embs = torch.nn.Embedding(3, 2)
    normal_vector_embs = torch.nn.Embedding(3, 2)
    projected_relation_embs = torch.nn.Embedding(3, 2)
    num_entities = 3
    num_relations = 3
    epislon = 1
    c = 0.5
    criterion = nn.MarginRankingLoss(margin=1, size_average=False)

    norm_of_entities = torch.norm(entities_embs.weight, p=2, dim=1)

    print("norm_of_entities shape: ", norm_of_entities.shape)
    print("norm_of_entities: ", norm_of_entities)
    print()


    square_norms_entities = torch.mul(norm_of_entities, norm_of_entities)

    print("square_norms_entities.shape: ", square_norms_entities.shape)
    print("square_norms_entities: ", square_norms_entities)
    print()


    entity_constraint = square_norms_entities - num_entities * 1.
    print("entity_constraint shape: ", entity_constraint.shape)
    print("entity_constraint: ", entity_constraint)
    print()

    entity_constraint = torch.sum(entity_constraint,dim=0)

    print("sum entity_constraint shape: ", entity_constraint.shape)
    print("sum entity_constraint: ", entity_constraint)
    print()

    print("normal_vector_embs shape: ", normal_vector_embs.weight.shape)
    print("normal_vector_embs: ", normal_vector_embs.weight.data)
    print()

    print("projected_relation_embs.weight shape: ", projected_relation_embs.weight.data.shape)
    print("projected_relation_embs.weight: ", projected_relation_embs.weight.data)
    print()

    orthogonalty_constraint_numerator = torch.mul(normal_vector_embs.weight,projected_relation_embs.weight)

    orthogonalty_constraint_numerator = torch.sum(orthogonalty_constraint_numerator,dim=1)


    orthogonalty_constraint_numerator = torch.mul(orthogonalty_constraint_numerator,
                                                  orthogonalty_constraint_numerator)

    print("shape orthogonalty_constraint_numerator: ", orthogonalty_constraint_numerator.shape)
    print("orthogonalty_constraint_numerator: ", orthogonalty_constraint_numerator)
    print()

    orthogonalty_constraint_denominator = torch.norm(projected_relation_embs.weight, p=2, dim=1)


    orthogonalty_constraint_denominator = torch.mul(orthogonalty_constraint_denominator,
                                                    orthogonalty_constraint_denominator)
    print("orthogonalty_constraint_denominator shape: ", orthogonalty_constraint_denominator.shape)
    print("orthogonalty_constraint_denominator: ", orthogonalty_constraint_denominator)
    print()


    relation_constraint = orthogonalty_constraint_numerator / orthogonalty_constraint_denominator
    print("relation_constraint shape: ", relation_constraint.shape)
    print("relation_constraint :", relation_constraint)
    print()

    relation_constraint = torch.sum(relation_constraint)

    relation_constraint = relation_constraint - num_relations * epislon
    soft_constraints = c * (entity_constraint + relation_constraint)

    print( soft_constraints)

    # loss = margin_ranking_loss + soft_constraints


# -*- coding: utf-8 -*-

import logging
import timeit

import numpy as np
import torch
import torch.optim as optim
from sklearn.utils import shuffle
from tqdm import tqdm

from pykeen.constants import *

log = logging.getLogger(__name__)


def _split_list_in_batches(input_list, batch_size):
    return [input_list[i:i + batch_size] for i in range(0, len(input_list), batch_size)]


def train_model(kg_embedding_model, all_entities, learning_rate, num_epochs, batch_size, pos_triples, device, seed):
    model_name = kg_embedding_model.model_name

    if model_name in [TRANS_E_NAME, TRANS_H_NAME, TRANS_D_NAME, TRANS_R_NAME, DISTMULT_NAME, UM_NAME, SE_NAME,
                      ERMLP_NAME, RESCAL_NAME]:
        return _train_basic_model(kg_embedding_model, all_entities, learning_rate, num_epochs, batch_size,
                                  pos_triples,
                                  device, seed)

    if model_name == CONV_E_NAME:
        return _train_conv_e_model(kg_embedding_model, all_entities, learning_rate, num_epochs, batch_size, pos_triples,
                                   device, seed)


def _train_basic_model(kg_embedding_model, all_entities, learning_rate, num_epochs, batch_size,
                       pos_triples, device,
                       seed):
    kg_embedding_model = kg_embedding_model.to(device)

    optimizer = optim.SGD(kg_embedding_model.parameters(), lr=learning_rate)

    loss_per_epoch = []

    log.info('****Run Model On %s****' % str(device).upper())

    num_pos_triples = pos_triples.shape[0]
    num_entities = all_entities.shape[0]

    start_training = timeit.default_timer()

    for epoch in tqdm(range(num_epochs)):
        np.random.seed(seed=seed)
        indices = np.arange(num_pos_triples)
        np.random.shuffle(indices)
        pos_triples = pos_triples[indices]
        start = timeit.default_timer()
        pos_batches = _split_list_in_batches(input_list=pos_triples, batch_size=batch_size)
        current_epoch_loss = 0.

        for i in range(len(pos_batches)):
            pos_batch = pos_batches[i]
            current_batch_size = len(pos_batch)
            batch_subjs = pos_batch[:, 0:1]
            batch_relations = pos_batch[:, 1:2]
            batch_objs = pos_batch[:, 2:3]

            num_subj_corrupt = len(pos_batch) // 2
            num_obj_corrupt = len(pos_batch) - num_subj_corrupt
            pos_batch = torch.tensor(pos_batch, dtype=torch.long, device=device)

            corrupted_subj_indices = np.random.choice(np.arange(0, num_entities), size=num_subj_corrupt)
            corrupted_subjects = np.reshape(all_entities[corrupted_subj_indices], newshape=(-1, 1))
            subject_based_corrupted_triples = np.concatenate(
                [corrupted_subjects, batch_relations[:num_subj_corrupt], batch_objs[:num_subj_corrupt]], axis=1)

            corrupted_obj_indices = np.random.choice(np.arange(0, num_entities), size=num_obj_corrupt)
            corrupted_objects = np.reshape(all_entities[corrupted_obj_indices], newshape=(-1, 1))

            object_based_corrupted_triples = np.concatenate(
                [batch_subjs[num_subj_corrupt:], batch_relations[num_subj_corrupt:], corrupted_objects], axis=1)

            neg_batch = np.concatenate([subject_based_corrupted_triples, object_based_corrupted_triples], axis=0)

            neg_batch = torch.tensor(neg_batch, dtype=torch.long, device=device)

            # Recall that torch *accumulates* gradients. Before passing in a
            # new instance, you need to zero out the gradients from the old instance
            optimizer.zero_grad()
            loss = kg_embedding_model(pos_batch, neg_batch)
            current_epoch_loss += (loss.item() * current_batch_size)

            loss.backward()
            optimizer.step()

        stop = timeit.default_timer()
        log.info("Epoch %s took %s seconds \n" % (str(epoch), str(round(stop - start))))
        # Track epoch loss
        loss_per_epoch.append(current_epoch_loss / len(pos_triples))

    stop_training = timeit.default_timer()
    log.info("Training took %s seconds \n" % (str(round(stop_training - start_training))))

    return kg_embedding_model, loss_per_epoch


def _train_conv_e_model(conv_e_model, all_entities, learning_rate, num_epochs, batch_size, pos_triples, device,
                        seed):
    conv_e_model = conv_e_model.to(device)

    optimizer = optim.Adam(conv_e_model.parameters(), lr=learning_rate)

    loss_per_epoch = []

    log.info('****Run Model On %s****' % str(device).upper())

    num_pos_triples = pos_triples.shape[0]
    num_entities = all_entities.shape[0]

    start_training = timeit.default_timer()

    for epoch in range(num_epochs):
        np.random.seed(seed=seed)
        indices = np.arange(num_pos_triples)
        np.random.shuffle(indices)
        pos_triples = pos_triples[indices]
        start = timeit.default_timer()
        num_positives = batch_size // 2
        # TODO: Make sure that batch = num_pos + num_negs
        num_negatives = batch_size - num_positives

        pos_batches = _split_list_in_batches(input_list=pos_triples, batch_size=num_positives)
        current_epoch_loss = 0.

        for i in range(len(pos_batches)):
            # TODO: Remove original subject and object from entity set
            pos_batch = pos_batches[i]
            current_batch_size = len(pos_batch)
            batch_subjs = pos_batch[:, 0:1]
            batch_relations = pos_batch[:, 1:2]
            batch_objs = pos_batch[:, 2:3]

            num_subj_corrupt = len(pos_batch) // 2
            num_obj_corrupt = len(pos_batch) - num_subj_corrupt
            pos_batch = torch.tensor(pos_batch, dtype=torch.long, device=device)

            corrupted_subj_indices = np.random.choice(np.arange(0, num_entities), size=num_subj_corrupt)
            corrupted_subjects = np.reshape(all_entities[corrupted_subj_indices], newshape=(-1, 1))
            subject_based_corrupted_triples = np.concatenate(
                [corrupted_subjects, batch_relations[:num_subj_corrupt], batch_objs[:num_subj_corrupt]], axis=1)

            corrupted_obj_indices = np.random.choice(np.arange(0, num_entities), size=num_obj_corrupt)
            corrupted_objects = np.reshape(all_entities[corrupted_obj_indices], newshape=(-1, 1))

            object_based_corrupted_triples = np.concatenate(
                [batch_subjs[num_subj_corrupt:], batch_relations[num_subj_corrupt:], corrupted_objects], axis=1)

            neg_batch = np.concatenate([subject_based_corrupted_triples, object_based_corrupted_triples], axis=0)

            neg_batch = torch.tensor(neg_batch, dtype=torch.long, device=device)

            batch = np.concatenate([pos_batch, neg_batch], axis=0)
            positive_labels = np.ones(shape=(current_batch_size))
            negative_labels = np.zeros(shape=(current_batch_size))
            labels = np.concatenate([positive_labels, negative_labels], axis=0)

            batch, labels = shuffle(batch, labels, random_state=seed)

            batch = torch.tensor(batch, dtype=torch.long)
            labels = torch.tensor(labels, dtype=torch.float)

            # Recall that torch *accumulates* gradients. Before passing in a
            # new instance, you need to zero out the gradients from the old
            # instance
            optimizer.zero_grad()
            loss = conv_e_model(batch, labels)
            current_epoch_loss += (loss.item() * current_batch_size)

            loss.backward()
            optimizer.step()

        stop = timeit.default_timer()
        log.info("Epoch %s took %s seconds \n" % (str(epoch), str(round(stop - start))))
        # Track epoch loss
        loss_per_epoch.append(current_epoch_loss / len(pos_triples))

    stop_training = timeit.default_timer()
    log.info("Training took %s seconds \n" % (str(round(stop_training - start_training))))

    return conv_e_model, loss_per_epoch



# -*- coding: utf-8 -*-
import logging
import timeit

import numpy as np
import torch

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


# def compute_mean_rank_and_hits_at_k(all_entities, kg_embedding_model, triples, device, k=10):
#     start = timeit.default_timer()
#     mean_rank, hits_at_k = _compute_metrics_new(all_entities=all_entities, kg_embedding_model=kg_embedding_model,
#                                                 triples=triples, device=device)
#
#     # ranks_subject_based, hits_at_k_subject_based = _compute_metrics(all_entities=all_entities,
#     #                                                                 kg_embedding_model=kg_embedding_model,
#     #                                                                 triples=triples, corrupt_suject=True, device=device,
#     #                                                                 k=k)
#     #
#     # ranks_object_based, hits_at_k_object_based = _compute_metrics(all_entities=all_entities,
#     #                                                               kg_embedding_model=kg_embedding_model,
#     #                                                               triples=triples, corrupt_suject=False, device=device,
#     #                                                               k=k)
#     # mean_rank = np.mean(ranks_subject_based + ranks_object_based)
#     #
#     # all_hits = hits_at_k_subject_based + hits_at_k_object_based
#     # hits_at_k = all_hits / (2. * triples.size)
#     #
#     # print("hits_at_k_subject_based: ", hits_at_k_subject_based / triples.size)
#     # print("hits_at_k_object_based: ", hits_at_k_object_based / triples.size)
#
#     stop = timeit.default_timer()
#     log.info("Evaluation took %s seconds \n" % (str(round(stop - start))))
#
#     return mean_rank, hits_at_k
#
#
# def compute_mean_rank(all_entities, kg_embedding_model, triples, device):
#     start = timeit.default_timer()
#     ranks_subject_based, _ = _compute_metrics(all_entities=all_entities, kg_embedding_model=kg_embedding_model,
#                                               triples=triples, corrupt_suject=True, device=device)
#
#     ranks_object_based, _ = _compute_metrics(all_entities=all_entities, kg_embedding_model=kg_embedding_model,
#                                              triples=triples, corrupt_suject=False, device=device)
#     ranks = ranks_subject_based + ranks_object_based
#     mean_rank = np.mean(ranks)
#
#     stop = timeit.default_timer()
#     log.info("Evaluation took %s seconds \n" % (str(round(stop - start))))
#
#     return mean_rank
#
#
# def compute_hits_at_k(all_entities, kg_embedding_model, triples, device, k=10):
#
#     _, hits_at_k_subject_based = _compute_metrics(all_entities=all_entities, kg_embedding_model=kg_embedding_model,
#                                                   triples=triples, corrupt_suject=True, k=k)
#
#     _, hits_at_k_object_based = _compute_metrics(all_entities=all_entities, kg_embedding_model=kg_embedding_model,
#                                                  triples=triples, corrupt_suject=False, k=k)
#
#     all_hits = hits_at_k_subject_based + hits_at_k_object_based
#     hits_at_k = all_hits / (2. * triples.size)
#     stop = timeit.default_timer()
#     log.info("Evaluation took %s seconds \n" % (str(round(stop - start))))
#
#     return hits_at_k
#
#
# def _compute_metrics(all_entities, kg_embedding_model, triples, corrupt_suject, device, k=10):
#     start = timeit.default_timer()
#     ranks = []
#     count_in_top_k = 0.
#
#     kg_embedding_model = kg_embedding_model.to(device)
#
#     column_to_maintain_offsets, corrupted_column_offsets, concatenate_fct = get_stratey_for_corrupting(
#         corrupt_suject=corrupt_suject)
#
#     start_of_columns_to_maintain, end_of_columns_to_maintain = column_to_maintain_offsets
#     start_of_corrupted_column, end_of_corrupted_column = corrupted_column_offsets
#
#     # Corrupt triples
#     for row_nmbr, row in enumerate(triples):
#
#         candidate_entities = np.delete(arr=all_entities,
#                                        obj=row[end_of_corrupted_column - 1:end_of_corrupted_column])
#
#         # Extract current test tuple: Either (subject,predicate) or (predicate,object)
#         tuple = np.reshape(a=triples[row_nmbr, start_of_columns_to_maintain:end_of_columns_to_maintain],
#                            newshape=(1, 2))
#         # Copy current test tuple
#         tuples = np.repeat(a=tuple, repeats=candidate_entities.shape[0], axis=0)
#
#         corrupted = concatenate_fct(candidate_entities=candidate_entities, tuples=tuples)
#         corrupted = torch.tensor(corrupted, dtype=torch.long, device=device)
#         scores_of_corrupted = kg_embedding_model.predict(corrupted)
#         pos_triple = np.array(row)
#         pos_triple = np.expand_dims(a=pos_triple, axis=0)
#         pos_triple = torch.tensor(pos_triple, dtype=torch.long, device=device)
#
#         score_of_positive = kg_embedding_model.predict(pos_triple)
#
#         scores = np.append(arr=scores_of_corrupted, values=score_of_positive)
#         indice_of_pos = scores.size - 1
#
#         sorted_score_indices = np.argsort(a=scores)
#
#         # Get index of first occurence that fulfills the condition
#         rank_of_positive = np.where(sorted_score_indices == indice_of_pos)[0][0]
#         ranks.append(rank_of_positive)
#
#         if rank_of_positive < k:
#             count_in_top_k += 1.
#
#     return ranks, count_in_top_k


def _compute_hits_at_k_new(hits_at_k_dict, rank_of_positive_subject_based, rank_of_positive_object_based):
    for k, value in hits_at_k_dict.items():
        if rank_of_positive_subject_based < k:
            value.append(1.)
        else:
            value.append(0.)

        if rank_of_positive_object_based < k:
            value.append(1.)
        else:
            value.append(0.)

    return hits_at_k_dict


def compute_metrics(all_entities, kg_embedding_model, triples, device):
    start = timeit.default_timer()
    ranks = []
    hits_at_k_dict = {k: [] for k in [1, 3, 5, 10]}

    kg_embedding_model = kg_embedding_model.to(device)

    # Corrupt triples
    for row_nmbr, row in enumerate(triples):
        candidate_entities_subject_based = np.delete(arr=all_entities, obj=row[0:1])
        candidate_entities_subject_based = np.reshape(candidate_entities_subject_based, newshape=(-1, 1))
        candidate_entities_object_based = np.delete(arr=all_entities, obj=row[2:3])
        candidate_entities_object_based = np.reshape(candidate_entities_object_based, newshape=(-1, 1))

        # Extract current test tuple: Either (subject,predicate) or (predicate,object)
        tuple_subject_based = np.reshape(a=triples[row_nmbr, 1:3], newshape=(1, 2))
        tuple_object_based = np.reshape(a=triples[row_nmbr, 0:2], newshape=(1, 2))

        # Copy current test tuple
        tuples_subject_based = np.repeat(a=tuple_subject_based, repeats=candidate_entities_subject_based.shape[0],
                                         axis=0)
        tuples_object_based = np.repeat(a=tuple_object_based, repeats=candidate_entities_object_based.shape[0], axis=0)

        corrupted_subject_based = np.concatenate([candidate_entities_subject_based, tuples_subject_based], axis=1)
        corrupted_subject_based = torch.tensor(corrupted_subject_based, dtype=torch.long, device=device)

        corrupted_object_based = np.concatenate([tuples_object_based, candidate_entities_object_based], axis=1)
        corrupted_object_based = torch.tensor(corrupted_object_based, dtype=torch.long, device=device)

        scores_of_corrupted_subjects = kg_embedding_model.predict(corrupted_subject_based)
        scores_of_corrupted_objects = kg_embedding_model.predict(corrupted_object_based)

        pos_triple = np.array(row)
        pos_triple = np.expand_dims(a=pos_triple, axis=0)
        pos_triple = torch.tensor(pos_triple, dtype=torch.long, device=device)

        score_of_positive = kg_embedding_model.predict(pos_triple)

        scores_subject_based = np.append(arr=scores_of_corrupted_subjects, values=score_of_positive)
        indice_of_pos_subject_based = scores_subject_based.size - 1

        scores_object_based = np.append(arr=scores_of_corrupted_objects, values=score_of_positive)
        indice_of_pos_object_based = scores_object_based.size - 1

        _, sorted_score_indices_subject_based = torch.sort(torch.tensor(scores_subject_based, dtype=torch.float),
                                                           descending=False)
        sorted_score_indices_subject_based = sorted_score_indices_subject_based.cpu().numpy()

        _, sorted_score_indices_object_based = torch.sort(torch.tensor(scores_object_based, dtype=torch.float),
                                                          descending=False)
        sorted_score_indices_object_based = sorted_score_indices_object_based.cpu().numpy()

        # Get index of first occurence that fulfills the condition
        rank_of_positive_subject_based = np.where(sorted_score_indices_subject_based == indice_of_pos_subject_based)[0][
            0]
        ranks.append(rank_of_positive_subject_based)

        rank_of_positive_object_based = np.where(sorted_score_indices_object_based == indice_of_pos_object_based)[0][0]
        ranks.append(rank_of_positive_object_based)

        # Compute hits@k for k in {1,3,5,10}
        hits_at_k_dict.update(
            _compute_hits_at_k_new(hits_at_k_dict, rank_of_positive_subject_based=rank_of_positive_subject_based,
                                   rank_of_positive_object_based=rank_of_positive_object_based))

    mean_rank = np.mean(ranks)

    for k, value in hits_at_k_dict.items():
        hits_at_k_dict[k] = np.mean(value)

    stop = timeit.default_timer()
    log.info("Evaluation took %s seconds \n" % (str(round(stop - start))))

    return mean_rank, hits_at_k_dict



# -*- coding: utf-8 -*-
from sklearn.model_selection import train_test_split

if __name__ == '__main__':
    ent = [1,20,2,3,30,4]
    train, test = train_test_split(ent,test_size = 0.33, random_state = 42)

    print(train)
    print(test)


# -*- coding: utf-8 -*-
import numpy as np

from evaluation_methods.abstract_evaluator import AbstractEvaluator
from keen.utilities import HITS_AT_K
from keen.utilities.evaluation_utils import get_stratey_for_corrupting


class HitsAtKEvaluator(AbstractEvaluator):
    METRIC = HITS_AT_K

    def _compute_hits_at_k(self, kg_embedding_model, data, k, corrupt_suject):
        num_triples = len(data)

        assert num_triples>=k

        column_to_maintain_offsets, corrupted_column_offsets, concatenate_fct = get_stratey_for_corrupting(
            corrupt_suject=corrupt_suject)

        start_of_columns_to_maintain, end_of_columns_to_maintain = column_to_maintain_offsets
        start_of_corrupted_colum, end_of_corrupted_column = corrupted_column_offsets

        scores_of_corrupted = []
        scores_of_originals = []

        # Corrupt triples
        for row in range(len(data)):
            tuple = np.reshape(a=data[row, start_of_columns_to_maintain:end_of_columns_to_maintain], newshape=(1, 2))
            tuples = np.repeat(a=tuple, repeats=num_triples - 1, axis=0)
            candidate_entities = np.delete(arr=data, obj=row, axis=0)[:,
                                 start_of_corrupted_colum:end_of_corrupted_column]

            corrupted = concatenate_fct(candidate_entities=candidate_entities, tuples=tuples)

            scores_of_corrupted.append(np.apply_along_axis(kg_embedding_model.predict, axis=1, arr=corrupted))
            scores_of_originals.append(kg_embedding_model.predict(data[row]))


        # Get top k
        scores_of_corrupted = np.concatenate(scores_of_corrupted,axis=-1)
        scores_of_originals = np.concatenate(scores_of_originals,axis=-1)
        all_scores = np.append(arr=scores_of_corrupted,values=scores_of_originals)
        all_scores = np.sort(a=all_scores)
        top_10_scores = all_scores[-k:]

        # TODO: Last step



        return None




# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

import json
import os
from collections import OrderedDict

import click

from pykeen.constants import (
    CONV_E_NAME, DISTMULT_NAME, ERMLP_NAME, EXECUTION_MODE, FILTER_NEG_TRIPLES, HPO_ITERS_ERROR_MSG,
    HPO_ITERS_PRINT_MSG, HPO_ITERS_PROMPT_MSG, HPO_MODE, NUM_OF_HPO_ITERS, OUTPUT_DIREC, PREFERRED_DEVICE, PYKEEN,
    RESCAL_NAME, SE_NAME, TEST_FILE_ERROR_MSG, TEST_FILE_PROMPT_MSG, TEST_SET_PATH, TEST_SET_RATIO,
    TRAINING_FILE_ERROR_MSG, TRAINING_FILE_PROMPT_MSG, TRAINING_MODE, TRAINING_SET_PATH, TRANS_D_NAME, TRANS_E_NAME,
    TRANS_H_NAME, TRANS_R_NAME, UM_NAME,
)

from pykeen.predict import start_predictions_piepline
from pykeen.run import run
from pykeen.utilities.cli_utils import (
    configure_distmult_training_pipeline, configure_ermlp_training_pipeline, configure_rescal_training_pipeline,
    configure_se_training_pipeline, configure_trans_d_training_pipeline, configure_trans_e_training_pipeline,
    configure_trans_h_training_pipeline, configure_trans_r_training_pipeline, configure_um_training_pipeline,
)
from pykeen.utilities.cli_utils.cli_print_msg_helper import (
    print_ask_for_evlauation_message, print_execution_mode_message, print_filter_negative_triples_message, print_intro,
    print_section_divider, print_test_ratio_message, print_test_set_message,
    print_training_set_message, print_welcome_message,
    print_model_selection_message, print_random_seed_message)
from pykeen.utilities.cli_utils.cli_query_helper import (
    ask_for_evaluation, ask_for_filtering_of_negatives, ask_for_test_set, get_input_path, query_output_directory,
    select_embedding_model, select_integer_value, select_keen_execution_mode, select_preferred_device,
    select_ratio_for_test_set,
)
from pykeen.utilities.cli_utils.conv_e_cli import configure_conv_e_hpo_pipeline, configure_conv_e_training_pipeline
from pykeen.utilities.cli_utils.distmult_cli import configure_distmult_hpo_pipeline
from pykeen.utilities.cli_utils.ermlp_cli import configure_ermlp_hpo_pipeline
from pykeen.utilities.cli_utils.rescal_cli import configure_rescal_hpo_pipeline
from pykeen.utilities.cli_utils.structured_embedding_cli import configure_se_hpo_pipeline
from pykeen.utilities.cli_utils.trans_d_cli import configure_trans_d_hpo_pipeline
from pykeen.utilities.cli_utils.trans_e_cli import configure_trans_e_hpo_pipeline
from pykeen.utilities.cli_utils.trans_h_cli import configure_trans_h_hpo_pipeline
from pykeen.utilities.cli_utils.trans_r_cli import configure_trans_r_hpo_pipeline
from pykeen.utilities.cli_utils.unstructured_model_cli import configure_um_hpo_pipeline
from pykeen.utilities.cli_utils.utils import summarize_results

MODEL_TRAINING_CONFIG_FUNCS = {
    TRANS_E_NAME: configure_trans_e_training_pipeline,
    TRANS_H_NAME: configure_trans_h_training_pipeline,
    TRANS_R_NAME: configure_trans_r_training_pipeline,
    TRANS_D_NAME: configure_trans_d_training_pipeline,
    SE_NAME: configure_se_training_pipeline,
    UM_NAME: configure_um_training_pipeline,
    DISTMULT_NAME: configure_distmult_training_pipeline,
    ERMLP_NAME: configure_ermlp_training_pipeline,
    RESCAL_NAME: configure_rescal_training_pipeline,
    CONV_E_NAME: configure_conv_e_training_pipeline
}

MODEL_HPO_CONFIG_FUNCS = {
    TRANS_E_NAME: configure_trans_e_hpo_pipeline,
    TRANS_H_NAME: configure_trans_h_hpo_pipeline,
    TRANS_R_NAME: configure_trans_r_hpo_pipeline,
    TRANS_D_NAME: configure_trans_d_hpo_pipeline,
    SE_NAME: configure_se_hpo_pipeline,
    UM_NAME: configure_um_hpo_pipeline,
    DISTMULT_NAME: configure_distmult_hpo_pipeline,
    ERMLP_NAME: configure_ermlp_hpo_pipeline,
    RESCAL_NAME: configure_rescal_hpo_pipeline,
    CONV_E_NAME: configure_conv_e_hpo_pipeline
}


def _configure_training_pipeline(model_name):
    model_config_func = MODEL_TRAINING_CONFIG_FUNCS.get(model_name)
    if model_config_func is None:
        raise KeyError(f'invalid model given: {model_name}')
    config = model_config_func(model_name)
    if config is None:
        raise NotImplementedError(f'{model_name} has not yet been implemented')
    return config


def _configure_hpo_pipeline(model_name):
    model_config_func = MODEL_HPO_CONFIG_FUNCS.get(model_name)
    if model_config_func is None:
        raise KeyError(f'invalid model given: {model_name}')
    config = model_config_func(model_name)
    if config is None:
        raise NotImplementedError(f'{model_name} has not yet been implemented')
    return config


def _configure_evaluation_specific_parameters(pykeen_exec_mode):
    """

    :param config:
    :return:
    """
    config = OrderedDict()

    if pykeen_exec_mode == TRAINING_MODE:
        # Step 1: Ask whether to evaluate the model
        print_ask_for_evlauation_message()
        is_evaluation_mode = ask_for_evaluation()
        print_section_divider()
    else:
        is_evaluation_mode = True

    # Step 2: Specify test set, if is_evaluation_mode==True
    if is_evaluation_mode:
        print_test_set_message()
        provide_test_set = ask_for_test_set()
        print_section_divider()

        if provide_test_set:
            test_set_path = get_input_path(prompt_msg=TEST_FILE_PROMPT_MSG,
                                           error_msg=TEST_FILE_ERROR_MSG)
            config[TEST_SET_PATH] = test_set_path
        else:
            print_test_ratio_message()
            test_set_ratio = select_ratio_for_test_set()
            config[TEST_SET_RATIO] = test_set_ratio

        print_section_divider()

        # Ask whether to use filtered negative triples
        print_filter_negative_triples_message()
        filter_negative_triples = ask_for_filtering_of_negatives()
        config[FILTER_NEG_TRIPLES] = filter_negative_triples
        print_section_divider()

    return config


def welcome_prompt():
    """

    :return:
    """
    # Step: Welcome + Intro
    print_welcome_message()
    print_section_divider()
    print_intro()
    print_section_divider()


def training_file_prompt(config):
    """

    :param config:
    :return:
    """
    print_training_set_message()
    config[TRAINING_SET_PATH] = get_input_path(
        prompt_msg=TRAINING_FILE_PROMPT_MSG,
        error_msg=TRAINING_FILE_ERROR_MSG,
    )
    return config


def execution_mode_prompt(config, lib_name=PYKEEN):
    """

    :param config:
    :return:
    """
    print_execution_mode_message()
    exec_mode = select_keen_execution_mode(lib_name=lib_name)
    config[EXECUTION_MODE] = exec_mode

    return config


def model_selection_prompt():
    """

    :return:
    """
    print_model_selection_message()
    model_name = select_embedding_model()
    print_section_divider()
    return model_name


def execution_mode_specific_prompt(config, model_name):
    """

    :param config:
    :param model_name:
    :return:
    """
    pykeen_exec_mode = config[EXECUTION_MODE]
    if pykeen_exec_mode == TRAINING_MODE:
        config.update(_configure_training_pipeline(model_name))
    elif pykeen_exec_mode == HPO_MODE:
        config.update(_configure_hpo_pipeline(model_name))

        # Query number of HPO iterations
        hpo_iter = select_integer_value(
            print_msg=HPO_ITERS_PRINT_MSG,
            prompt_msg=HPO_ITERS_PROMPT_MSG,
            error_msg=HPO_ITERS_ERROR_MSG)
        config[NUM_OF_HPO_ITERS] = hpo_iter
        print_section_divider()
    return config


def random_seed_prompt(config):
    """Query random seed."""

    print_random_seed_message()
    config[SEED] = select_integer_value(print_msg=SEED_PRINT_MSG,
                                        prompt_msg=SEED_PROMPT_MSG,
                                        error_msg=SEED_ERROR_MSG)

    return config


def device_prompt(config):
    """

    :param config:
    :return:
    """
    config[PREFERRED_DEVICE] = select_preferred_device()
    return config


def output_direc_prompt(config):
    """

    :param config:
    :return:
    """
    config[OUTPUT_DIREC] = query_output_directory()
    return config


def prompt_config():
    """

    :return:
    """
    config = OrderedDict()

    # Step 1: Welcome + Intro
    welcome_prompt()

    # Step 2: Ask for training file
    config = training_file_prompt(config=config)
    print_section_divider()

    # Step 3: Ask for execution mode
    config = execution_mode_prompt(config=config)
    print_section_divider()

    # Step 4: Ask for model
    model_name = model_selection_prompt()
    print_section_divider()

    # Step 5: Query parameters depending on the selected execution mode
    config = execution_mode_specific_prompt(config=config, model_name=model_name)
    print_section_divider()

    config.update(_configure_evaluation_specific_parameters(config[EXECUTION_MODE]))

    # Step 6: Please select a random seed
    config = random_seed_prompt(config=config)
    print_section_divider()

    # Step 7: Query device to train on
    config = device_prompt(config=config)
    print_section_divider()

    # Step 8: Define output directory
    config = output_direc_prompt(config=config)
    print_section_divider()

    return config


@click.command()
@click.option('-c', '--config', type=click.File(), help='A PyKEEN JSON configuration file')
def main(config):
    """PyKEEN: A software for training and evaluating knowledge graph embeddings."""
    if config is not None:
        config = json.load(config)
    else:
        config = prompt_config()

    run(config)


@click.command()
@click.option('-m', '--model-directory', type=click.Path(file_okay=False, dir_okay=True))
@click.option('-d', '--data-directory', type=click.Path(file_okay=False, dir_okay=True))
def predict(model_directory: str, data_directory: str):
    """Predict new links based on trained model."""
    start_predictions_piepline(model_directory, data_directory)


@click.command()
@click.option('-d', '--directory', type=click.Path(file_okay=False, dir_okay=True), default=os.getcwd())
@click.option('-o', '--output', type=click.File('w'))
def summarize(directory: str, output):
    """Summarize contents of training and evaluation."""
    summarize_results(directory, output)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import numpy as np
import logging
import timeit
from evaluation_methods.abstract_evaluator import AbstractEvaluator
from keen.utilities import MEAN_RANK
from keen.utilities.evaluation_utils.evaluation_helper import get_stratey_for_corrupting

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

class MeanRankEvaluator(AbstractEvaluator):
    METRIC = MEAN_RANK

    def _compute_ranks(self, kg_embedding_model, triples, corrupt_suject):
        start = timeit.default_timer()
        num_triples = len(triples)
        ranks = []
        in_top_k = []

        column_to_maintain_offsets, corrupted_column_offsets, concatenate_fct = get_stratey_for_corrupting(
            corrupt_suject=corrupt_suject)

        start_of_columns_to_maintain, end_of_columns_to_maintain = column_to_maintain_offsets
        start_of_corrupted_colum, end_of_corrupted_column = corrupted_column_offsets

        # Corrupt triples
        for row in range(len(triples)):
            tuple = np.reshape(a=triples[row, start_of_columns_to_maintain:end_of_columns_to_maintain], newshape=(1, 2))
            tuples = np.repeat(a=tuple, repeats=num_triples - 1, axis=0)
            candidate_entities = np.delete(arr=triples, obj=row, axis=0)[:,
                                 start_of_corrupted_colum:end_of_corrupted_column]

            corrupted = concatenate_fct(candidate_entities=candidate_entities, tuples=tuples)
            scores_of_corrupted = kg_embedding_model.predict(corrupted)
            pos_triple = np.array(triples[row])
            pos_triple= np.expand_dims(a=pos_triple,axis=0)

            score_of_original = kg_embedding_model.predict(pos_triple)
            scores = np.append(arr=scores_of_corrupted, values=score_of_original)
            scores = np.sort(a=scores)
            # Get index of first occurence that fulfills the condition
            ranks.append(np.where(scores == score_of_original)[0][0])

            k = 10
            top_k = scores[-k:]
            if score_of_original in top_k:
                in_top_k.append(1.)

        stop = timeit.default_timer()
        log.info("Evaluation took %s seconds \n" % (str(round(stop - start))))

        return ranks

    def start_evaluation(self, test_data, kg_embedding_model):
        """

        :param test_data:
        :param kg_embedding_model:
        :return:
        """

        ranks = self._compute_ranks(kg_embedding_model=kg_embedding_model, triples=test_data, corrupt_suject=True)
        ranks += self._compute_ranks(kg_embedding_model=kg_embedding_model, triples=test_data, corrupt_suject=False)
        mean_rank = np.mean(ranks)

        return mean_rank, MEAN_RANK



# -*- coding: utf-8 -*-
import os
import sys

w_dir = os.path.dirname(os.getcwd())
sys.path.append(w_dir)

import click
import yaml

from keen.utilities.pipeline import Pipeline


@click.command()
@click.option('-cfg_path', help='path to config file', required=True)
def main(cfg_path):
    with open(cfg_path, 'r') as ymlfile:
        cfg = yaml.load(ymlfile)


    pipeline = Pipeline(config=cfg, seed=2)

    trained_model, eval_summary, entity_to_embedding, relation_to_embedding = pipeline.start_training()

    print(eval_summary)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod

__all__ = ['AbstractHPOptimizer']


class AbstractHPOptimizer(ABC):

    @abstractmethod
    def optimize_hyperparams(self, config, path_to_kg, device, seed):
        pass




import pickle

if __name__ == '__main__':

    data_path = '/Users/mehdi/PycharmProjects/kg_embeddings_pipeline/data/corpora/out/25-09-2018_09:29:12/entities_to_embeddings.pkl'
    with open(data_path, 'rb') as f:
        data = pickle.load(f)


    print(data)


# -*- coding: utf-8 -*-

import torch
import torch.autograd
import torch.nn as nn

from keen.constants import *


class GenericEmbeddings(nn.Module):

    def __init__(self, config):
        super(GenericEmbeddings, self).__init__()
        # A simple lookup table that stores embeddings of a fixed dictionary and size

        num_entities = config[NUM_ENTITIES]
        num_relations = config[NUM_RELATIONS]
        self.embedding_dim = config[EMBEDDING_DIM]
        margin_loss = config[MARGIN_LOSS]

        self.entities_embeddings = nn.Embedding(num_entities, self.embedding_dim)
        self.relation_embeddings = nn.Embedding(num_relations, self.embedding_dim)
        self.R_k_s_embeddings = nn.Embedding(num_relations, self.embedding_dim * self.embedding_dim)
        self.alpha_k_s = nn.Embedding(num_relations, 1)
        self.W_k_s = nn.Embedding(num_relations, self.embedding_dim * self.embedding_dim)
        self.margin_loss = margin_loss
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=False)
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    def forward(self, pos_exmpls, neg_exmpls):
        pos_heads = pos_exmpls[:, 0:1]
        pos_relations = pos_exmpls[:, 1:2]
        pos_tails = pos_exmpls[:, 2:3]

        neg_heads = neg_exmpls[:, 0:1]
        neg_relations = neg_exmpls[:, 1:2]
        neg_tails = neg_exmpls[:, 2:3]

        pos_h_embs = self.entities_embeddings(pos_heads)
        pos_r_embs = self.relation_embeddings(pos_relations)
        pos_t_embs = self.entities_embeddings(pos_tails)

        neg_h_embs = self.entities_embeddings(neg_heads)
        neg_r_embs = self.relation_embeddings(neg_relations)
        neg_t_embs = self.entities_embeddings(neg_tails)

        R_k_s = self.self.R_k_s_embeddings(pos_relations).view(-1, self.embedding_dim,
                                                               self.embedding_dim)
        W_k_s = self.self.W_k_s(pos_relations).view(-1, self.embedding_dim,
                                                    self.embedding_dim)
        alpha_k_s = self.alpha_k_s(pos_relations)

        # Project entities into relation space
        proj_pos_heads = self._project_entities(pos_h_embs, W_k_s)
        proj_pos_tails = self._project_entities(pos_t_embs, W_k_s)

        proj_neg_heads = self._project_entities(neg_h_embs, W_k_s)
        proj_neg_tails = self._project_entities(neg_t_embs, W_k_s)

        pos_score = self.compute_scores(projected_head=proj_pos_heads, projected_tail=proj_pos_tails, R_k=R_k_s,
                                        alpha_k=alpha_k_s)
        neg_score = self.compute_scores(projected_head=proj_neg_heads, projected_tail=proj_neg_tails, R_k=R_k_s,
                                        alpha_k=alpha_k_s)

        loss = self.compute_loss(pos_score=pos_score, neg_score=neg_score)

        return loss

    def _project_entities(self, head_emb, W_k):
        projected_head = torch.mm(W_k, head_emb)

        return projected_head

    def compute_scores(self, projected_head, projected_tail, R_k, alpha_k):
        """

        :param projected_head:
        :param r_embs:
        :param projected_tail:
        :return:
        """
        # TODO: - torch.abs(h_emb + r_emb - t_emb)
        # Compute score and transform result to 1D tensor
        # TODO: Score is the negative of the distance
        first_part = torch.mm(R_k, projected_head)
        second_part = alpha_k * projected_tail
        score = - torch.sum(torch.abs(first_part - second_part))

        return score




import numpy as np
import torch

if __name__ == '__main__':
    scores = np.array([10, 1, 4, 2])
    sorted_score_indices= np.argsort(scores)
    indice_of_pos = 3
    print(sorted_score_indices)
    rank_of_positive = np.where(sorted_score_indices == indice_of_pos)[0][0]
    print(rank_of_positive)
    print(sorted_score_indices[:2])

    print("####torch based sorting#######")
    sorted_scores, sorted_score_indices = torch.sort(torch.tensor(scores,dtype=torch.float),descending=False)
    print(sorted_score_indices)
    print(sorted_scores)
    exit(0)


    pos_triples = np.array([[0,1,2],[0,3,0],[0,1,5]])
    subj_rels = pos_triples[:,0:2]
    labels = []
    entities = np.arange(0,5)
    entities = np.expand_dims(entities,axis=-1)

    y1 = np.array([[1, 2,2], [1, 3,2], [1, 2,2], [2, 2,2]])
    z = np.array([1, 2,2])
    # print((y1 == z))
    # print((y1 == z).all(axis=1)*1.)

    for subj_rel in subj_rels:
        subj_rel = np.expand_dims(subj_rel,axis=0)
        subj_rel_rep = np.repeat(subj_rel,axis=0,repeats=entities.size)
        candidates = np.concatenate([subj_rel_rep,entities],axis=-1)
        label_for_t= []
        for candi in candidates:
            label = (candi == pos_triples).all(axis=1).any()* 1.
            label_for_t.append(label)

        print(label_for_t)
        labels.append(label_for_t)

    labels = np.array(labels)
    print(labels.shape)
    print(labels)




import numpy as np


if __name__ == '__main__':
    a = np.ones(shape=(2,2))
    outfile = '/Users/mehdi/PycharmProjects/kg_embeddings_pipeline/data/test_serialized_data/matrix.npy'
    np.save(outfile, a)

    print(np.load(outfile))


# -*- coding: utf-8 -*-

"""Implementation of TransD."""

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from pykeen.constants import *

__all__ = ['TransD']

"""
TODO: Check, whether it makes sense to integrate identity matrices.
"""


class TransD(nn.Module):

    def __init__(self, config):
        super(TransD, self).__init__()
        self.model_name = TRANS_D_NAME
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.entity_embedding_dim = config[EMBEDDING_DIM]
        self.relation_embedding_dim = self.entity_embedding_dim
        self.margin_loss = config[MARGIN_LOSS]
        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.entity_embeddings = nn.Embedding(self.num_entities, self.entity_embedding_dim, max_norm=1)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.relation_embedding_dim, max_norm=1)
        self.entity_projections = nn.Embedding(self.num_entities, self.entity_embedding_dim)
        self.relation_projections = nn.Embedding(self.num_relations, self.relation_embedding_dim)

        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)
        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]
        # self._initialize()

    def _compute_scores(self, h_embs, r_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = torch.mul(distances, distances)

        return distances

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float, device=self.device)
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _project_entities(self, entity_embs, entity_proj_vecs, relation_projs):
        entity_embs = entity_embs
        relation_projs = relation_projs.unsqueeze(-1)
        entity_proj_vecs = entity_proj_vecs.unsqueeze(-1).permute([0, 2, 1])

        transfer_matrices = torch.matmul(relation_projs, entity_proj_vecs)

        projected_entity_embs = torch.einsum('nmk,nk->nm', [transfer_matrices, entity_embs])

        return projected_entity_embs

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        h_embs = self.entity_embeddings(heads).view(-1, self.entity_embedding_dim)
        r_embs = self.relation_embeddings(relations).view(-1, self.relation_embedding_dim)
        t_embs = self.entity_embeddings(tails).view(-1, self.entity_embedding_dim)

        h_proj_vec_embs = self.entity_projections(heads).view(-1, self.entity_embedding_dim)
        r_projs_embs = self.relation_projections(relations).view(-1, self.relation_embedding_dim)
        t_proj_vec_embs = self.entity_projections(tails).view(-1, self.entity_embedding_dim)

        proj_heads = self._project_entities(h_embs, h_proj_vec_embs, r_projs_embs)
        proj_tails = self._project_entities(t_embs, t_proj_vec_embs, r_projs_embs)

        scores = self._compute_scores(h_embs=proj_heads, r_embs=r_embs, t_embs=proj_tails)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.entity_embedding_dim)
        pos_r_embs = self.relation_embeddings(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.entity_embedding_dim)

        pos_h_proj_vec_embs = self.entity_projections(pos_heads).view(-1, self.entity_embedding_dim)
        pos_r_projs_embs = self.relation_projections(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_proj_vec_embs = self.entity_projections(pos_tails).view(-1, self.entity_embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.entity_embedding_dim)
        neg_r_embs = self.relation_embeddings(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.entity_embedding_dim)

        neg_h_proj_vec_embs = self.entity_projections(neg_heads).view(-1, self.entity_embedding_dim)
        neg_r_projs_embs = self.relation_projections(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_proj_vec_embs = self.entity_projections(neg_tails).view(-1, self.entity_embedding_dim)

        # Project entities
        proj_pos_heads = self._project_entities(pos_h_embs, pos_h_proj_vec_embs, pos_r_projs_embs)
        proj_pos_tails = self._project_entities(pos_t_embs, pos_t_proj_vec_embs, pos_r_projs_embs)

        proj_neg_heads = self._project_entities(neg_h_embs, neg_h_proj_vec_embs, neg_r_projs_embs)
        proj_neg_tails = self._project_entities(neg_t_embs, neg_t_proj_vec_embs, neg_r_projs_embs)

        pos_scores = self._compute_scores(h_embs=proj_pos_heads, r_embs=pos_r_embs, t_embs=proj_pos_tails)
        neg_scores = self._compute_scores(h_embs=proj_neg_heads, r_embs=neg_r_embs, t_embs=proj_neg_tails)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss



from keen.hyper_parameter_optimizer.random_search_optimizer import RandomSearchHPO
from keen.utilities import CLASS_NAME


def get_hyper_parameter_optimizer(config, evaluator):
    class_name = config[CLASS_NAME]
    if class_name == 'RandomSearchHPO':
        return RandomSearchHPO(evaluator=evaluator)




import json
import urllib

api_key = open(".api_key").read()
query = '/m/06rf7'
service_url = 'https://www.googleapis.com/freebase/v1/search'
params = {
        'query': query,
        'key': api_key
}
url = service_url + '?' + urllib.urlencode(params)
response = json.loads(urllib.urlopen(url).read())
for result in response['result']:
    print(result['name'] + ' (' + str(result['score']) + ')')


# -*- coding: utf-8 -*-

import random

import numpy as np

from pykeen.constants import *
from pykeen.hyper_parameter_optimizer.abstract_hyper_params_optimizer import AbstractHPOptimizer
from pykeen.utilities.evaluation_utils.metrics_computations import compute_metrics
from pykeen.utilities.initialization_utils.module_initialization_utils import get_kg_embedding_model
from pykeen.utilities.train_utils import train_model

__all__ = ['RandomSearchHPO']


class RandomSearchHPO(AbstractHPOptimizer):

    def _sample_conv_e_params(self, hyparams_dict):
        kg_model_config = OrderedDict()
        # Sample params which are dependent on each other
        embedding_dimensions = hyparams_dict[EMBEDDING_DIM]
        sampled_index = random.choice(range(len(embedding_dimensions)))
        kg_model_config[EMBEDDING_DIM] = embedding_dimensions[sampled_index]
        kg_model_config[CONV_E_HEIGHT] = hyparams_dict[CONV_E_HEIGHT][sampled_index]
        kg_model_config[CONV_E_WIDTH] = hyparams_dict[CONV_E_WIDTH][sampled_index]
        kg_model_config[CONV_E_KERNEL_HEIGHT] = hyparams_dict[CONV_E_KERNEL_HEIGHT][sampled_index]
        kg_model_config[CONV_E_KERNEL_WIDTH] = hyparams_dict[CONV_E_KERNEL_WIDTH][sampled_index]

        del hyparams_dict[EMBEDDING_DIM]
        del hyparams_dict[CONV_E_HEIGHT]
        del hyparams_dict[CONV_E_WIDTH]
        del hyparams_dict[CONV_E_KERNEL_HEIGHT]
        del hyparams_dict[CONV_E_KERNEL_WIDTH]

        kg_model_config.update(self._sample_params(hyparams_dict))

        return kg_model_config

    def _sample_params(self, hyperparams_dict):
        kg_model_config = OrderedDict()
        for param, values in hyperparams_dict.items():
            if isinstance(values, list):
                kg_model_config[param] = random.choice(values)
            else:
                kg_model_config[param] = values

        return kg_model_config

    def optimize_hyperparams(self, mapped_train_tripels, mapped_test_tripels, entity_to_id, rel_to_id, config,
                             device, seed=None):
        if seed is not None:
            np.random.seed(seed=seed)

        trained_models = []
        eval_results = []
        entity_to_ids = []
        rel_to_ids = []
        models_params = []
        eval_summaries = []
        epoch_losses = []

        config = config.copy()
        max_iters = config[NUM_OF_HPO_ITERS]

        sample_fct = self._sample_conv_e_params if config[
                                                       KG_EMBEDDING_MODEL_NAME] == CONV_E_NAME else self._sample_params

        for _ in range(max_iters):
            eval_summary = OrderedDict()

            # Sample hyper-params
            kg_embedding_model_config = sample_fct(config)
            kg_embedding_model_config[NUM_ENTITIES] = len(entity_to_id)
            kg_embedding_model_config[NUM_RELATIONS] = len(rel_to_id)
            kg_embedding_model_config[SEED] = seed

            # Configure defined model
            kg_embedding_model = get_kg_embedding_model(config=kg_embedding_model_config)

            models_params.append(kg_embedding_model_config)
            entity_to_ids.append(entity_to_id)
            rel_to_ids.append(rel_to_id)

            all_entities = np.array(list(entity_to_id.values()))

            trained_model, epoch_loss = train_model(
                kg_embedding_model=kg_embedding_model,
                all_entities=all_entities,
                learning_rate=kg_embedding_model_config[LEARNING_RATE],
                num_epochs=kg_embedding_model_config[NUM_EPOCHS],
                batch_size=kg_embedding_model_config[BATCH_SIZE],
                pos_triples=mapped_train_tripels,
                device=device,
                seed=seed
            )

            # Evaluate trained model
            mean_rank, hits_at_k = compute_metrics(
                all_entities=all_entities,
                kg_embedding_model=trained_model,
                mapped_train_triples=mapped_train_tripels,
                mapped_test_triples=mapped_test_tripels,
                device=device
            )

            # TODO: Define HPO metric
            eval_summary[MEAN_RANK] = mean_rank
            eval_summary[HITS_AT_K] = hits_at_k
            eval_results.append(hits_at_k[10])
            eval_summaries.append(eval_summary)
            trained_models.append(trained_model)
            epoch_losses.append(epoch_loss)

        index_of_max = np.argmax(a=eval_results)

        return trained_models[index_of_max], \
               epoch_losses[index_of_max], \
               entity_to_ids[index_of_max], \
               rel_to_ids[index_of_max], \
               eval_summaries[index_of_max], \
               models_params[index_of_max]

    @staticmethod
    def run(mapped_train_tripels, mapped_test_tripels, entity_to_id, rel_to_id, config, device, seed):
        hpo = RandomSearchHPO()
        return hpo.optimize_hyperparams(
            mapped_train_tripels=mapped_train_tripels,
            mapped_test_tripels=mapped_test_tripels,
            entity_to_id=entity_to_id,
            rel_to_id=rel_to_id,
            config=config,
            device=device,
            seed=seed
        )



import json
if __name__ == '__main__':
    path = '/Users/mehdi/.kg_embeddings_pipeline/csqa_wiki_data/temp.json'
    with open(path, 'r', encoding='utf-8') as fp:
        data = json.load(fp)

    subjects = data.keys()
    predicates = data.values()
    print(subjects)
    print()
    print(predicates)
    print()

    for p in predicates:
        print(p)
        print()

    print(len(subjects))
    print(len(predicates))


# -*- coding: utf-8 -*-
import json
import logging
import multiprocessing
import os

from keen.utilities import KG_EMBEDDINGS_PIPELINE_DIR, CSQA_WIKIDATA

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
from corpus_reader.AbstractReader import AbstractReader



class CSQAWikiDataReader(AbstractReader):

    def _extract(self, argument_tuple):
        """

        :param argument_tuple:
        :return:
        """
        triples = []
        keys_of_kg_dict, kg = argument_tuple
        for subject in keys_of_kg_dict:
            value_dict = kg[subject]

            for predicate, object_list in value_dict.items():
                if not object_list:
                    continue
                for object in object_list:
                    triples.append((subject, predicate, object))

        return triples

    def _split_list_in_chunks(self, input_list, num_chunks):
        return [input_list[i::num_chunks] for i in range(num_chunks)]

    def transorm_corpus_to_id_format(self, temp_corpus, corpus_path):
        """

        :param temp_corpus:
        :param corpus_path:
        :return:
        """

        log.info("---------Load Corpus From Disk---------")
        with open(temp_corpus, 'r', encoding='utf-8') as fp:
            data = json.load(fp)

        keys = list(data.keys())
        num_processes = multiprocessing.cpu_count()
        num_processes = 10
        log.info("Number of Used CPUs: ", num_processes)

        chunk_keys = self._split_list_in_chunks(input_list=keys, num_chunks=num_processes)
        chunksize = len(chunk_keys[0])

        log.info("---------Create Triples---------")
        triples = self._extract((data.keys(),data))

        # with Pool(num_processes) as p:
        #     # triple_lists = p.map(self._extract, [(subset_keys, data) for subset_keys in chunk_keys])
        #     triple_lists = p.imap(self._extract, [(subset_keys, data) for subset_keys in chunk_keys],chunksize=chunksize)
        #
        # triples = [item for sublist in triple_lists for item in sublist]

        log.info("---------Write Corpus---------")
        with open(corpus_path, 'w', encoding='utf-8') as f1:
            f1.write('@Comment@ Subject Predicate Object\n')
            for triple in triples:
                f1.write('\t'.join(triple) + '\n')

        os.remove(temp_corpus)

    def retreive_knowledge_graph(self):
        """

        :return:
        """
        data_dir = os.path.join(KG_EMBEDDINGS_PIPELINE_DIR, CSQA_WIKIDATA)
        os.makedirs(data_dir, exist_ok=True)

        corpus_path = os.path.join(data_dir, "csqa_wikidata_short_1.csv")

        if not os.path.exists(corpus_path):
            temp_path = os.path.join(data_dir, "wikidata_short_1.json")
            self.transorm_corpus_to_id_format(temp_corpus=temp_path, corpus_path=corpus_path)

        return corpus_path



# -*- coding: utf-8 -*-

"""Script for starting the pipeline and saving the results."""

import json
import os
import pickle
import time
from typing import Mapping, Optional

import torch

from pykeen.constants import (
    ENTITY_TO_EMBEDDING, ENTITY_TO_ID, EVAL_SUMMARY, LOSSES, OUTPUT_DIREC, RELATION_TO_EMBEDDING, RELATION_TO_ID,
    TRAINED_MODEL,
)
from pykeen.utilities.pipeline import Pipeline


def run(config: Mapping,
        output_directory: Optional[str] = None,
        training_path: Optional[str] = None):
    """Run PyKEEN using a given configuration."""
    if output_directory is None:
        output_directory = os.path.join(config[OUTPUT_DIREC], time.strftime("%Y-%m-%d_%H:%M:%S"))
    os.makedirs(output_directory, exist_ok=True)

    pipeline = Pipeline(config=config)

    pipeline_outcome, params = pipeline.start(path_to_train_data=training_path)

    out_path = os.path.join(output_directory, 'configuration.json')
    with open(out_path, 'w') as handle:
        json.dump(params, handle, indent=2)

    out_path = os.path.join(output_directory, 'entities_to_embeddings.pkl')
    with open(out_path, 'wb') as handle:
        pickle.dump(pipeline_outcome[ENTITY_TO_EMBEDDING], handle, protocol=pickle.HIGHEST_PROTOCOL)

    out_path = os.path.join(output_directory, 'relations_to_embeddings.pkl')
    with open(out_path, 'wb') as handle:
        pickle.dump(pipeline_outcome[RELATION_TO_EMBEDDING], handle, protocol=pickle.HIGHEST_PROTOCOL)

    out_path = os.path.join(output_directory, 'entity_to_id.json')
    with open(out_path, 'w') as handle:
        json.dump(pipeline_outcome[ENTITY_TO_ID], handle, indent=2)

    out_path = os.path.join(output_directory, 'relation_to_id.json')
    with open(out_path, 'w') as handle:
        json.dump(pipeline_outcome[RELATION_TO_ID], handle, indent=2)

    out_path = os.path.join(output_directory, 'losses.json')
    with open(out_path, 'w') as handle:
        json.dump(pipeline_outcome[LOSSES], handle, indent=2)

    eval_summary = pipeline_outcome.get(EVAL_SUMMARY)
    if eval_summary is not None:
        out_path = os.path.join(output_directory, 'evaluation_summary.json')
        with open(out_path, 'w') as handle:
            json.dump(eval_summary, handle, indent=2)

    # Save trained model
    out_path = os.path.join(output_directory, 'trained_model.pkl')
    torch.save(pipeline_outcome[TRAINED_MODEL].state_dict(), out_path)



# -*- coding: utf-8 -*-

"""Setup.py for pykeen."""

import codecs
import os
import re

import setuptools

MODULE = 'pykeen'
PACKAGES = setuptools.find_packages(where='src')
META_PATH = os.path.join('src', MODULE, '__init__.py')
KEYWORDS = ['Knowledge Graph Embeddings', 'Machine Learning', 'Data Mining', 'Linked Data']
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Chemistry',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Information Analysis',
]
INSTALL_REQUIRES = [
    'numpy; python_version == "3.6"',
    'numpy; python_version == "3.7"',
    'scikit-learn==0.19.1; python_version == "3.6"',
    'scikit-learn; python_version == "3.7"',
    'scipy',
    'click==6.7',
    'torch==0.4.0; python_version == "3.6"',
    'torch==0.4.1; python_version == "3.7"',
    'torchvision==0.2.1',
    'prompt_toolkit',
    'tqdm',
    'pandas',
    'rdflib',
]
EXTRAS_REQUIRE = {
    'docs': [
        'sphinx',
        'sphinx-rtd-theme',
        'sphinx-click',
    ],
}
ENTRY_POINTS = {
    'console_scripts': [
        'pykeen = pykeen.cli:main',
        'pykeen-summarize = pykeen.cli:summarize',
        'pykeen-predict = pykeen.cli:predict',
    ]
}

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """Build an absolute path from *parts* and return the contents of the resulting file. Assume UTF-8 encoding."""
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    meta_match = re.search(
        r'^__{meta}__ = ["\']([^"\']*)["\']'.format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string'.format(meta=meta))


def get_long_description():
    """Get the long_description from the README.rst file. Assume UTF-8 encoding."""
    with codecs.open(os.path.join(HERE, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


if __name__ == '__main__':
    setuptools.setup(
        name=find_meta('title'),
        version=find_meta('version'),
        description=find_meta('description'),
        long_description=get_long_description(),
        url=find_meta('url'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        maintainer=find_meta('author'),
        maintainer_email=find_meta('email'),
        license=find_meta('license'),
        classifiers=CLASSIFIERS,
        keywords=KEYWORDS,
        packages=PACKAGES,
        package_dir={'': 'src'},
        include_package_data=True,
        install_requires=INSTALL_REQUIRES,
        entry_points=ENTRY_POINTS,
        zip_safe=False,
    )



# -*- coding: utf-8 -*-

"""Implementation of TransR."""

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from pykeen.constants import *

__all__ = ['TransR']

"""
Constraints: 
 * ||h||_2 <= 1: Done
 * ||r||_2 <= 1: Done
 * ||t||_2 <= 1: Done
 * ||h*M_r||_2 <= 1: Done
 * ||t*M_r||_2 <= 1: Done
"""


class TransR(nn.Module):

    def __init__(self, config):
        super(TransR, self).__init__()
        self.model_name = TRANS_R_NAME
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.entity_embedding_dim = config[EMBEDDING_DIM]
        self.relation_embedding_dim = config[RELATION_EMBEDDING_DIM]
        self.margin_loss = config[MARGIN_LOSS]
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)
        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        # max_norm = 1 according to the paper
        # TODO: max_norm < 1.
        self.entity_embeddings = nn.Embedding(self.num_entities, self.entity_embedding_dim, norm_type=2, max_norm=1)
        # max_norm = 1 according to the paper
        self.relation_embeddings = nn.Embedding(self.num_relations, self.relation_embedding_dim, norm_type=2,
                                                max_norm=1)
        self.projection_matrix_embs = nn.Embedding(self.num_relations,
                                                   self.relation_embedding_dim * self.entity_embedding_dim)
        self.scoring_fct_norm = config[SCORING_FUNCTION_NORM]
        self._initialize()

    def _compute_scores(self, h_embs, r_embs, t_embs):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = torch.mul(distances, distances)

        return distances

    def _compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float, device=self.device)
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def _project_entities(self, entity_embs, projection_matrix_embs):
        """

        :param entity_embs:
        :param projection_matrix_embs:
        :return:
        """
        projected_entity_embs = torch.einsum('nk,nkd->nd', [entity_embs, projection_matrix_embs])

        projected_entity_embs = torch.clamp(projected_entity_embs, max=1.)

        return projected_entity_embs

    # TODO: Initilaize relation matrices as identiy matrices
    def _initialize(self):
        lower_bound = -6 / np.sqrt(self.entity_embedding_dim)
        upper_bound = 6 / np.sqrt(self.entity_embedding_dim)
        nn.init.uniform_(self.entity_embeddings.weight.data, a=lower_bound, b=upper_bound)
        nn.init.uniform_(self.relation_embeddings.weight.data, a=lower_bound, b=upper_bound)

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads).view(-1, self.entity_embedding_dim)
        relation_embs = self.relation_embeddings(relations).view(-1, self.relation_embedding_dim)
        tail_embs = self.entity_embeddings(tails).view(-1, self.entity_embedding_dim)

        proj_matrix_embs = self.projection_matrix_embs(relations).view(-1, self.entity_embedding_dim,
                                                                       self.relation_embedding_dim)

        proj_heads_embs = self._project_entities(head_embs, proj_matrix_embs)
        proj_tails_embs = self._project_entities(tail_embs, proj_matrix_embs)

        scores = self._compute_scores(h_embs=proj_heads_embs, r_embs=relation_embs, t_embs=proj_tails_embs)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.entity_embedding_dim)
        pos_r_embs = self.relation_embeddings(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.entity_embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.entity_embedding_dim)
        neg_r_embs = self.relation_embeddings(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.entity_embedding_dim)

        proj_matrix_embs = self.projection_matrix_embs(pos_relations).view(-1, self.entity_embedding_dim,
                                                                           self.relation_embedding_dim)

        # Project entities into relation space
        proj_pos_heads_embs = self._project_entities(pos_h_embs, proj_matrix_embs)
        proj_pos_tails_embs = self._project_entities(pos_t_embs, proj_matrix_embs)

        proj_neg_heads_embs = self._project_entities(neg_h_embs, proj_matrix_embs)
        proj_neg_tails_embs = self._project_entities(neg_t_embs, proj_matrix_embs)

        pos_scores = self._compute_scores(h_embs=proj_pos_heads_embs, r_embs=pos_r_embs, t_embs=proj_pos_tails_embs)
        neg_scores = self._compute_scores(h_embs=proj_neg_heads_embs, r_embs=neg_r_embs, t_embs=proj_neg_tails_embs)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch
import torch.autograd
import torch.nn as nn

from keen.constants import *

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class RotE(nn.Module):

    def __init__(self, config):
        super(RotE, self).__init__()
        self.model_name = ROT_E
        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]
        self.embedding_dim = config[EMBEDDING_DIM]
        margin_loss = config[MARGIN_LOSS]

        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() and config[PREFERRED_DEVICE] == GPU else CPU)

        self.entities_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)
        self.R_k_s_embeddings = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)
        self.alpha_k_s = nn.Embedding(self.num_relations, 1)
        self.W_k_s = nn.Embedding(self.num_relations, self.embedding_dim * self.embedding_dim)

        self.margin_loss = margin_loss
        self.criterion = nn.MarginRankingLoss(margin=self.margin_loss, size_average=True)

        # TODO: FIXME
        self._init()

    def _init(self):
        lower_bound = -6 / np.sqrt(self.embedding_dim)
        upper_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(self.entities_embeddings.weight.data, a=lower_bound, b=upper_bound)
        nn.init.uniform_(self.relation_embeddings.weight.data, a=lower_bound, b=upper_bound)

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def compute_loss(self, pos_scores, neg_scores):
        """

        :param pos_scores:
        :param neg_scores:
        :return:
        """

        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float, device=self.device)
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)

        return loss

    def compute_score(self, proj_head_embs, proj_tail_embs, R_k, alpha_k):
        """

        :param h_embs:
        :param r_embs:
        :param t_embs:
        :return:
        """

        # Apply rotation
        rotated_pos_head_embs = torch.mm(R_k, proj_head_embs)
        scaled_tail_embs = alpha_k * proj_tail_embs
        # Add the vector element wise
        sum_res = rotated_pos_head_embs - scaled_tail_embs
        distances = torch.norm(sum_res, dim=1).view(size=(-1,))

        return distances

    def predict(self, triples):
        """

        :param head:
        :param relation:
        :param tail:
        :return:
        """
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)

        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entities_embeddings(heads).view(-1, self.embedding_dim)
        relation_embs = self.relation_embeddings(relations).view(-1, self.embedding_dim)
        tail_embs = self.entities_embeddings(tails).view(-1, self.embedding_dim)

        scores = self.compute_score(h_embs=head_embs, r_embs=relation_embs, t_embs=tail_embs)

        return scores.detach().cpu().numpy()

    def forward(self, pos_exmpls, neg_exmpls):
        """

        :param pos_exmpls:
        :param neg_exmpls:
        :return:
        """

        # Normalise embeddings of entities
        norms = torch.norm(self.entities_embeddings.weight, p=2, dim=1).data
        self.entities_embeddings.weight.data = self.entities_embeddings.weight.data.div(
            norms.view(self.num_entities, 1).expand_as(self.entities_embeddings.weight))

        pos_heads = pos_exmpls[:, 0:1]
        pos_relations = pos_exmpls[:, 1:2]
        pos_tails = pos_exmpls[:, 2:3]

        neg_heads = neg_exmpls[:, 0:1]
        neg_relations = neg_exmpls[:, 1:2]
        neg_tails = neg_exmpls[:, 2:3]

        pos_h_embs = self.entities_embeddings(pos_heads)

        pos_t_embs = self.entities_embeddings(pos_tails).view(-1, self.embedding_dim)

        neg_h_embs = self.entities_embeddings(neg_heads).view(-1, self.embedding_dim)

        neg_t_embs = self.entities_embeddings(neg_tails).view(-1, self.embedding_dim)

        R_k_s = self.R_k_s_embeddings(pos_relations).view(-1, self.embedding_dim,
                                                          self.embedding_dim)
        W_k_s = self.W_k_s(pos_relations).view(-1, self.embedding_dim,
                                               self.embedding_dim)

        print("W_k_s shape: ", W_k_s.shape)
        print("pos_h_embs shape: ", pos_h_embs.t().shape)
        exit(0)
        alpha_k_s = self.alpha_k_s(pos_relations)

        # Project entities into relation space
        proj_pos_heads = torch.mm(W_k_s, pos_h_embs)
        proj_pos_tails = torch.mm(W_k_s, pos_t_embs)

        proj_neg_heads = torch.mm(W_k_s, neg_h_embs)
        proj_neg_tails = torch.mm(W_k_s, neg_t_embs)

        pos_scores = self.compute_score(projected_head=proj_pos_heads, projected_tail=proj_pos_tails, R_k=R_k_s,
                                        alpha_k=alpha_k_s)
        neg_scores = self.compute_score(projected_head=proj_neg_heads, projected_tail=proj_neg_tails, R_k=R_k_s,
                                        alpha_k=alpha_k_s)

        loss = self.compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)

        return loss




import torch
import numpy as np
if __name__ == '__main__':
    # a = torch.tensor(np.array([1,1,1],dtype=np.float32))
    # b = torch.transpose(torch.tensor(np.ones(shape=(3,4),dtype=np.float32)),0,1)

    # c = a*b

    a = torch.tensor([[3,3,3],[2,2,2]], dtype=torch.float)
    b = torch.tensor([[1,1,1],[1,1,1]], dtype=torch.float)
    c = torch.tensor([[1,1],[2,2]], dtype=torch.long)

    print(a-b)

    s = torch.sum(a-b,1)
    square= torch.mul(s,s)
    rt = torch.sqrt(square)
    dist = -rt
    print(s)
    print(square)
    print(rt)
    print(dist)

    print(torch.abs(s))

    # print(a-b)


    # print(torch.dist(a+b,c))
    # print(torch.norm(a-b,1))
    # a = np.array([1,5,0],dtype=np.int)
    # print(np.sort(a))
    # print(c)




# -*- coding: utf-8 -*-

"""Utilities for the command line interface."""

from pykeen.utilities.cli_utils.distmult_cli import configure_distmult_training_pipeline
from pykeen.utilities.cli_utils.ermlp_cli import configure_ermlp_training_pipeline
from pykeen.utilities.cli_utils.rescal_cli import configure_rescal_training_pipeline
from pykeen.utilities.cli_utils.structured_embedding_cli import configure_se_training_pipeline
from pykeen.utilities.cli_utils.trans_d_cli import configure_trans_d_training_pipeline
from pykeen.utilities.cli_utils.trans_e_cli import configure_trans_e_training_pipeline
from pykeen.utilities.cli_utils.trans_h_cli import configure_trans_h_training_pipeline
from pykeen.utilities.cli_utils.trans_r_cli import configure_trans_r_training_pipeline
from pykeen.utilities.cli_utils.unstructured_model_cli import configure_um_training_pipeline
